from gersemi.formatted_file import FormattedFile
from gersemi.return_codes import FAIL, SUCCESS
from gersemi.task_result import TaskResult
from gersemi.utils import fromfile


def check_formatting(formatted_file: FormattedFile) -> TaskResult:
    if formatted_file.before != formatted_file.after:
        return TaskResult(
            path=formatted_file.path,
            return_code=FAIL,
            to_stderr=f"{fromfile(formatted_file.path)} would be reformatted",
        )
    return TaskResult(path=formatted_file.path, return_code=SUCCESS)


def quiet_check_formatting(formatted_file: FormattedFile) -> TaskResult:
    if formatted_file.before != formatted_file.after:
        return TaskResult(path=formatted_file.path, return_code=FAIL)
    return TaskResult(path=formatted_file.path, return_code=SUCCESS)



from dataclasses import astuple
from functools import partial
from itertools import chain
import multiprocessing as mp
import multiprocessing.dummy as mp_dummy
from pathlib import Path
import sys
from typing import Callable, Dict, Iterable, Tuple
from gersemi.cache import create_cache
from gersemi.configuration import Configuration
from gersemi.custom_command_definition_finder import find_custom_command_definitions
from gersemi.formatter import create_formatter, Formatter
from gersemi.mode import Mode
from gersemi.parser import PARSER as parser
from gersemi.result import Result, Error, apply, get_error_message
from gersemi.return_codes import SUCCESS, INTERNAL_ERROR
from gersemi.task_result import TaskResult
from gersemi.tasks.check_formatting import check_formatting, quiet_check_formatting
from gersemi.tasks.forward_to_stdout import forward_to_stdout
from gersemi.tasks.format_file import format_file, FormattedFile
from gersemi.tasks.rewrite_in_place import rewrite_in_place
from gersemi.tasks.show_diff import show_colorized_diff, show_diff
from gersemi.utils import smart_open
from gersemi.keywords import Keywords

CHUNKSIZE = 16


print_to_stdout = partial(print, file=sys.stdout, end="")
print_to_stderr = partial(print, file=sys.stderr)


def get_files(paths: Iterable[Path]) -> Iterable[Path]:
    def get_files_from_single_path(path):
        if path.is_dir():
            return chain(path.rglob("CMakeLists.txt"), path.rglob("*.cmake"))
        return [path]

    return set(
        item.resolve() if item != Path("-") else item
        for path in paths
        for item in get_files_from_single_path(path)
    )


def has_custom_command_definition(code: str) -> bool:
    lowercased = code.lower()
    has_function_definition = "function" in lowercased and "endfunction" in lowercased
    has_macro_definition = "macro" in lowercased and "endmacro" in lowercased
    return has_function_definition or has_macro_definition


def find_custom_command_definitions_in_file_impl(filepath: Path) -> Dict[str, Keywords]:
    with smart_open(filepath, "r") as f:
        code = f.read()
    if not has_custom_command_definition(code):
        return {}

    parse_tree = parser.parse(code)
    return find_custom_command_definitions(parse_tree)


def find_custom_command_definitions_in_file(
    filepath: Path,
) -> Result[Dict[str, Keywords]]:
    return apply(find_custom_command_definitions_in_file_impl, filepath)


def find_all_custom_command_definitions(
    paths: Iterable[Path], pool
) -> Dict[str, Keywords]:
    result = {}

    files = get_files(paths)
    find = find_custom_command_definitions_in_file

    for defs in pool.imap_unordered(find, files, chunksize=CHUNKSIZE):
        if isinstance(defs, Error):
            print_to_stderr(get_error_message(defs))
        else:
            result.update(defs)
    return result


def select_task(mode: Mode, configuration: Configuration):
    return {
        Mode.ForwardToStdout: lambda _: forward_to_stdout,
        Mode.RewriteInPlace: lambda _: rewrite_in_place,
        Mode.CheckFormatting: lambda config: quiet_check_formatting
        if config.quiet
        else check_formatting,
        Mode.ShowDiff: lambda config: show_colorized_diff
        if config.color
        else show_diff,
    }[mode](configuration)


def run_task(
    path: Path, formatter: Formatter, task: Callable[[FormattedFile], TaskResult]
) -> TaskResult:
    formatted_file: Result[FormattedFile] = apply(format_file, path, formatter)
    if isinstance(formatted_file, Error):
        return TaskResult(
            path=path,
            return_code=INTERNAL_ERROR,
            to_stderr=get_error_message(formatted_file),
        )
    return task(formatted_file)


def consume_task_result(task_result: TaskResult) -> Tuple[Path, int]:
    path, return_code, to_stdout, to_stderr = astuple(task_result)
    if to_stdout != "":
        print_to_stdout(to_stdout)
    if to_stderr != "":
        print_to_stderr(to_stderr)
    return path, return_code


def create_pool(is_stdin_in_sources):
    if is_stdin_in_sources:
        return mp_dummy.Pool
    return partial(mp.Pool, processes=mp.cpu_count())


def filter_already_formatted_files(
    cache, configuration_summary: str, files: Iterable[Path]
):
    known_files = cache.get_files(configuration_summary)
    for f in files:
        if f not in known_files:
            yield f
        else:
            s = f.stat()
            if (s.st_size, s.st_mtime_ns) != known_files[f]:
                yield f


def store_files_in_cache(
    mode: Mode, cache, configuration_summary: str, files: Iterable[Path]
) -> None:
    if mode in [Mode.CheckFormatting, Mode.RewriteInPlace]:
        cache.store_files(configuration_summary, files)


def run(mode: Mode, configuration: Configuration, sources: Iterable[Path]):
    configuration_summary = configuration.summary()
    requested_files = get_files(sources)
    task = select_task(mode, configuration)

    pool_cm = create_pool(Path("-") in requested_files)
    with create_cache() as cache, pool_cm() as pool:
        files_to_format = list(
            filter_already_formatted_files(
                cache, configuration_summary, requested_files
            )
        )
        custom_command_definitions = find_all_custom_command_definitions(
            set(configuration.definitions), pool
        )
        formatter = create_formatter(
            not configuration.unsafe,
            configuration.line_length,
            custom_command_definitions,
        )
        execute = partial(run_task, formatter=formatter, task=task)

        results = [
            consume_task_result(result)
            for result in pool.imap_unordered(
                execute, files_to_format, chunksize=CHUNKSIZE
            )
        ]
        store_files_in_cache(
            mode,
            cache,
            configuration_summary,
            (path for path, code in results if code == SUCCESS and path != Path("-")),
        )

        return max((code for _, code in results), default=SUCCESS)



from itertools import dropwhile, takewhile
from gersemi.ast_helpers import is_keyword
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper


def can_be_formatted_into_single_line(children):
    return len(just_values(children)) <= 4


def just_values(children):
    is_not_keyword = lambda child: not is_cache(child) and not is_parent_scope(child)
    return list(takewhile(is_not_keyword, children[1:]))


is_parent_scope = is_keyword("PARENT_SCOPE")
is_cache = is_keyword("CACHE")
is_force = is_keyword("FORCE")


def is_cache_type(children):
    return any(map(is_cache, children))


class Set(BaseCommandInvocationDumper):
    def _format_cache_part(self, children):
        result = self._try_to_format_into_single_line(children, separator=" ")
        if result is not None:
            return result

        cache, cache_type, docstring = map(self.visit, children)
        return f"{cache} {cache_type.lstrip()}\n{docstring}"

    def _format_name_and_values(self, children):
        if can_be_formatted_into_single_line(children):
            result = self._try_to_format_into_single_line(children, separator=" ")
            if result is not None:
                return result

        return "\n".join(map(self.visit, children))

    def _cache_arguments(self, tree):
        if is_force(tree.children[-1]):
            *begin, force = tree.children
        else:
            begin, force = tree.children, None
        is_not_cache = lambda child: not is_cache(child)
        name_and_values = [*takewhile(is_not_cache, begin)]
        cache_part = [*dropwhile(is_not_cache, begin)]

        formatted_begin = self._format_name_and_values(name_and_values)
        formatted_cache_part = self._format_cache_part(cache_part)
        formatted_force = "" if force is None else "\n" + self.visit(force)
        return f"{formatted_begin}\n{formatted_cache_part}{formatted_force}"

    def complex_argument(self, tree):
        arguments = tree.children[0]
        if len(arguments.children) <= 4:
            result = self._try_to_format_into_single_line(
                arguments.children, separator=" ", prefix="(", postfix=")"
            )
            if result is not None:
                return result

        with self.indented():
            formatted_arguments = self.visit_children(arguments)
        return "\n".join([self._indent("("), *formatted_arguments, self._indent(")")])

    def arguments(self, tree):
        if can_be_formatted_into_single_line(tree.children):
            result = self._try_to_format_into_single_line(tree.children, separator=" ")
            if result is not None:
                return result

        if is_cache_type(tree.children):
            return self._cache_arguments(tree)

        return "\n".join(self.visit_children(tree))



from gersemi.formatted_file import FormattedFile
from gersemi.return_codes import SUCCESS
from gersemi.task_result import TaskResult
from gersemi.utils import smart_open


def rewrite_in_place(formatted_file: FormattedFile) -> TaskResult:
    if formatted_file.before != formatted_file.after:
        with smart_open(
            formatted_file.path, "w", newline=formatted_file.newlines_style
        ) as f:
            print(formatted_file.after, file=f, end="")
    return TaskResult(path=formatted_file.path, return_code=SUCCESS)



from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
import sys
from typing import Any, Iterator, List


def pop_all(in_list: List) -> List:
    popped, in_list[:] = in_list[:], []
    return popped


def advance(iterator: Iterator, times: int, default: Any) -> Any:
    result = default
    for _ in range(times):
        new_result = next(iterator, default)
        if new_result == default:
            break
        result = new_result
    return result


def fromfile(path):
    return "<stdin>" if path == Path("-") else str(path)


def tofile(path):
    return "<stdout>" if path == Path("-") else str(path)


class StdinWrapper:
    @staticmethod
    @lru_cache(maxsize=None)
    def read():
        return sys.stdin.read()


def standard_stream_open(mode):
    if mode is None or mode == "" or "r" in mode:
        return StdinWrapper()
    return sys.stdout


@contextmanager
def smart_open(filename, mode, *args, **kwargs):
    if filename == Path("-"):
        yield standard_stream_open(mode)
    else:
        try:
            fh = open(filename, mode, *args, **kwargs)
            yield fh
        finally:
            fh.close()



from gersemi.dumper import Dumper
from gersemi.parser import BARE_PARSER, PARSER
from gersemi.sanity_checker import check_code_equivalence


def noop(*_):
    pass


class Formatter:
    def __init__(self, sanity_checker, line_length: int, custom_command_definitions):
        self.sanity_checker = sanity_checker
        self.line_length = line_length
        self.custom_command_definitions = custom_command_definitions

    def format(self, code):
        dumper = Dumper(self.line_length, self.custom_command_definitions)
        result = dumper.visit(PARSER.parse(code))
        self.sanity_checker(BARE_PARSER, code, result)
        return result


def create_formatter(do_sanity_check, line_length, custom_command_definitions):
    sanity_checker = check_code_equivalence if do_sanity_check else noop
    return Formatter(sanity_checker, line_length, custom_command_definitions)



from lark import Token, Tree
from lark.visitors import Transformer
from gersemi.ast_helpers import is_newline
from gersemi.types import Nodes


class DowncaseIdentifiers(Transformer):
    def IDENTIFIER(self, token):
        return Token(
            "IDENTIFIER",
            token.lower(),
            pos_in_stream=token.pos_in_stream,
            line=token.line,
            column=token.column,
            end_line=token.end_line,
            end_column=token.end_column,
        )

    def else_term(self, token):
        return self.IDENTIFIER(token[0])

    elseif = else_term
    endforeach = else_term
    endfunction = else_term
    endif = else_term
    endmacro = else_term
    endwhile = else_term
    foreach = else_term
    function = else_term
    if_term = else_term
    macro = else_term
    while_term = else_term


class FlattenBracketComment(Transformer):
    def bracket_comment(self, children) -> Tree:
        *_, bracket_argument = children
        return Tree("bracket_comment", bracket_argument.children)


class CleanUpComplexArgument(Transformer):
    def complex_argument(self, children):
        return Tree("complex_argument", children[1:-1])


class CleanUpNewlines(Transformer):
    def newline_or_gap(self, children):
        return Token("NEWLINE", "".join(children)[:2])

    def arguments(self, children: Nodes) -> Tree:
        return Tree("arguments", [child for child in children if not is_newline(child)])


class ParsingTransformer(
    DowncaseIdentifiers, FlattenBracketComment, CleanUpComplexArgument, CleanUpNewlines
):
    pass



from lark import Discard
from lark.visitors import Interpreter, Transformer
from gersemi.ast_helpers import is_keyword
from gersemi.keywords import Keywords


class IgnoreThisDefinition:
    pass


class DropIrrelevantElements(Transformer):
    def _discard(self, _):
        raise Discard

    def non_command_element(self, children):
        if len(children) == 1 and isinstance(children[0], IgnoreThisDefinition):
            return children[0]
        raise Discard

    def line_comment(self, children):
        if len(children) > 0 and children[0].strip() == "gersemi: ignore":
            return IgnoreThisDefinition()
        raise Discard

    NEWLINE = _discard
    bracket_comment = _discard


is_parse_argv = is_keyword("PARSE_ARGV")


class CMakeInterpreter(Interpreter):
    def __init__(self, stack=None):
        self.stack = {} if stack is None else stack
        self.found_commands = {}

    @property
    def _inner_scope(self):
        return type(self)(stack=self.stack.copy())

    def _eval_variables(self, arg):
        for name, value in self.stack.items():
            arg = arg.replace(f"${{{name}}}", ";".join(value))
        return arg.split(";")

    def _set(self, arguments):
        name, *values = self.visit_children(arguments)
        self.stack[name] = [
            item for value in values for item in self._eval_variables(value)
        ]

    def _new_command(self, arguments):
        if len(arguments.children) > 0:
            name = arguments.children[0]
            return self.visit(name)
        raise RuntimeError

    def _cmake_parse_arguments(self, arguments):
        if is_parse_argv(arguments.children[0]):
            keywords = arguments.children[3:6]
        else:
            keywords = arguments.children[1:4]

        options, one_value_keywords, multi_value_keywords = [
            self._eval_variables(self.visit(item)) for item in keywords
        ]
        return Keywords(options, one_value_keywords, multi_value_keywords)

    def start(self, tree):
        return self.visit(tree.children[0])

    def file(self, tree):
        self.visit_children(tree)
        return self.found_commands

    def _should_definition_be_ignored(self, block):
        _, *maybe_body, _ = block.children
        if maybe_body:
            body, *_ = maybe_body
            return len(body.children) > 0 and isinstance(
                body.children[0], IgnoreThisDefinition
            )
        return False

    def block(self, tree):
        if self._should_definition_be_ignored(tree):
            return

        subinterpreter = self._inner_scope
        block_begin, *body, _ = subinterpreter.visit_children(tree)
        name, *_ = block_begin
        if name is None:
            return

        keywords, *_ = body
        if len(keywords) > 0:
            self.found_commands[name.lower()] = keywords[0]
        else:
            self.found_commands[name.lower()] = Keywords()

    def block_body(self, tree):
        return [
            item for item in self.visit_children(tree) if isinstance(item, Keywords)
        ][:1]

    def command_invocation(self, tree):
        identifier, arguments = tree.children
        command_interpreters = {
            "cmake_parse_arguments": self._cmake_parse_arguments,
            "function": self._new_command,
            "macro": self._new_command,
            "set": self._set,
        }
        return command_interpreters.get(identifier, lambda *args: None)(arguments)

    def _join(self, tree):
        return "".join(self.visit_children(tree))

    def complex_argument(self, tree):
        return f"({self.visit_children(tree)})"

    def _extract_first(self, tree):
        return tree.children[0]

    def quoted_argument(self, tree):
        return tree.children[0] if tree.children else ""

    bracket_argument = _join
    commented_argument = _join
    unquoted_argument = _extract_first


def find_custom_command_definitions(tree):
    tree = DropIrrelevantElements(visit_tokens=True).transform(tree)
    return CMakeInterpreter().visit(tree)



from gersemi.command_line_formatter import CommandLineFormatter
from gersemi.keyword_with_pairs_formatter import KeywordWithPairsFormatter
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class AndroidAddTestData(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = [
        "FILES_DEST",
        "LIBS_DEST",
        "DEVICE_OBJECT_STORE",
        "DEVICE_TEST_DIR",
    ]
    multi_value_keywords = ["FILES", "LIBS", "NO_LINK_REGEX"]


class CheckCSourceCompiles(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["FAIL_REGEX"]


class CheckCXXSourceCompiles(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["FAIL_REGEX"]


class CheckFortranSourceCompiles(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["SRC_EXT"]
    multi_value_keywords = ["FAIL_REGEX"]


class CheckFortranSourceRuns(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["SRC_EXT"]


class CheckIPOSupported(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["RESULT", "OUTPUT"]
    multi_value_keywords = ["LANGUAGES"]


class CheckIncludeFiles(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["LANGUAGE"]


class CheckOBJCSourceCompiles(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["FAIL_REGEX"]


class CheckOBJCXXSourceCompiles(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["FAIL_REGEX"]


class CheckPIESupported(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["OUTPUT_VARIABLE"]
    multi_value_keywords = ["LANGUAGES"]


class CheckStructHasMember(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["LANGUAGE"]


class CheckTypeSize(ArgumentAwareCommandInvocationDumper):
    options = ["BUILTIN_TYPES_ONLY"]
    one_value_keywords = ["LANGUAGE"]


class CMakeAddFortranSubdirectory(
    CommandLineFormatter, ArgumentAwareCommandInvocationDumper
):
    options = ["NO_EXTERNAL_INSTALL", "LINK_LIBRARIES"]
    one_value_keywords = ["PROJECT", "ARCHIVE_DIR", "RUNTIME_DIR"]
    multi_value_keywords = ["LIBRARIES", "LINK_LIBS", "CMAKE_COMMAND_LINE"]
    keyword_formatters = {"CMAKE_COMMAND_LINE": "_format_command_line"}


class CMakePrintProprties(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = [
        "TARGETS",
        "SOURCES",
        "DIRECTORIES",
        "TESTS",
        "CACHE_ENTRIES",
        "PROPERTIES",
    ]


class ConfigurePackageConfigFile(ArgumentAwareCommandInvocationDumper):
    options = ["NO_SET_AND_CHECK_MACRO", "NO_CHECK_REQUIRED_COMPONENTS_MACRO"]
    one_value_keywords = ["INSTALL_DESTINATION", "INSTALL_PREFIX"]
    multi_value_keywords = ["PATH_VARS"]


class CPackAddComponent(ArgumentAwareCommandInvocationDumper):
    options = ["HIDDEN", "REQUIRED", "DISABLED", "DOWNLOADED"]
    one_value_keywords = [
        "DISPLAY_NAME",
        "DESCRIPTION",
        "GROUP",
        "ARCHIVE_FILE",
        "PLIST",
    ]
    multi_value_keywords = ["DEPENDS", "INSTALL_TYPES"]


class CPackAddComponentGroup(ArgumentAwareCommandInvocationDumper):
    options = ["EXPANDED", "BOLD_TITLE"]
    one_value_keywords = ["DISPLAY_NAME", "DESCRIPTION", "PARENT_GROUP"]


class CPackAddInstallType(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DISPLAY_NAME"]


class CPackConfigureDownloads(ArgumentAwareCommandInvocationDumper):
    options = ["ALL", "ADD_REMOVE", "NO_ADD_REMOVE"]
    one_value_keywords = ["UPLOAD_DIRECTORY"]


class CPackIFWAddRepository(ArgumentAwareCommandInvocationDumper):
    options = ["DISABLED"]
    one_value_keywords = ["URL", "USERNAME", "PASSWORD", "DISPLAY_NAME"]


class CPackIFWConfigureComponent(ArgumentAwareCommandInvocationDumper):
    options = [
        "COMMON",
        "ESSENTIAL",
        "VIRTUAL",
        "FORCED_INSTALLATION",
        "REQUIRES_ADMIN_RIGHTS",
    ]
    one_value_keywords = [
        "NAME",
        "DISPLAY_NAME",
        "DESCRIPTION",
        "UPDATE_TEXT",
        "VERSION",
        "RELEASE_DATE",
        "SCRIPT",
        "PRIORITY",
        "SORTING_PRIORITY",
        "CHECKABLE",
        "DEFAULT",
    ]
    multi_value_keywords = [
        "DEPENDS",
        "DEPENDENCIES",
        "AUTO_DEPEND_ON",
        "LICENSES",
        "USER_INTERFACES",
        "TRANSLATIONS",
        "REPLACES",
    ]


class CPackIFWConfigureComponentGroup(ArgumentAwareCommandInvocationDumper):
    options = [
        "VIRTUAL",
        "FORCED_INSTALLATION",
        "REQUIRES_ADMIN_RIGHTS",
    ]

    one_value_keywords = [
        "NAME",
        "DISPLAY_NAME",
        "DESCRIPTION",
        "UPDATE_TEXT",
        "VERSION",
        "RELEASE_DATE",
        "SCRIPT",
        "PRIORITY",
        "SORTING_PRIORITY",
        "DEFAULT",
        "CHECKABLE",
    ]
    multi_value_keywords = [
        "DEPENDS",
        "DEPENDENCIES",
        "AUTO_DEPEND_ON",
        "LICENSES",
        "USER_INTERFACES",
        "TRANSLATIONS",
        "REPLACES",
    ]


class CPackIFWUpdateRepository(ArgumentAwareCommandInvocationDumper):
    options = ["ADD", "REMOVE", "REPLACE", "DISABLED"]
    one_value_keywords = [
        "URL",
        "OLD_URL",
        "NEW_URL",
        "USERNAME",
        "PASSWORD",
        "DISPLAY_NAME",
    ]


class CTestCoverageCollectGCOV(
    CommandLineFormatter, ArgumentAwareCommandInvocationDumper
):
    options = ["GLOB", "DELETE", "QUIET"]
    one_value_keywords = ["TARBALL", "SOURCE", "BUILD", "GCOV_COMMAND"]
    multi_value_keywords = ["GCOV_OPTIONS"]
    keyword_formatters = {"GCOV_OPTIONS": "_format_command_line"}


class ExternalProjectAdd(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    one_value_keywords = [
        # Directory
        "PREFIX",
        "TMP_DIR",
        "STAMP_DIR",
        "LOG_DIR",
        "DOWNLOAD_DIR",
        "SOURCE_DIR",
        "BINARY_DIR",
        "INSTALL_DIR",
        # Download Step
        "URL_HASH",
        "URL_MD5",
        "DOWNLOAD_NAME",
        "DOWNLOAD_NO_EXTRACT",
        "TIMEOUT",
        "HTTP_USERNAME",
        "HTTP_PASSWORD",
        "TLS_VERIFY",
        "TLS_CAINFO",
        "NETRC",
        "NETRC_FILE",
        "GIT_REPOSITORY",
        "GIT_TAG",
        "GIT_REMOTE_NAME",
        "GIT_SUBMODULES_RECURSE",
        "GIT_SHALLOW",
        "GIT_PROGRESS",
        "GIT_REMOTE_UPDATE_STRATEGY",
        "SVN_REPOSITORY",
        "SVN_REVISION",
        "SVN_USERNAME",
        "SVN_PASSWORD",
        "SVN_TRUST_CERT",
        "HG_REPOSITORY",
        "HG_TAG",
        "CVS_REPOSITORY",
        "CVS_MODULE",
        "CVS_TAG",
        # Update/Patch Step
        "UPDATE_DISCONNECTED",
        # Configure Step
        "CMAKE_COMMAND",
        "CMAKE_GENERATOR",
        "CMAKE_GENERATOR_PLATFORM",
        "CMAKE_GENERATOR_TOOLSET",
        "CMAKE_GENERATOR_INSTANCE",
        "SOURCE_SUBDIR",
        "CONFIGURE_HANDLED_BY_BUILD",
        # Build Step
        "BUILD_IN_SOURCE",
        "BUILD_ALWAYS",
        # Install Step
        # Test Step
        "TEST_BEFORE_INSTALL",
        "TEST_AFTER_INSTALL",
        "TEST_EXCLUDE_FROM_MAIN",
        # Output Logging
        "LOG_DOWNLOAD",
        "LOG_UPDATE",
        "LOG_PATCH",
        "LOG_CONFIGURE",
        "LOG_BUILD",
        "LOG_INSTALL",
        "LOG_TEST",
        "LOG_MERGED_STDOUTERR",
        "LOG_OUTPUT_ON_FAILURE",
        # Terminal Access
        "USES_TERMINAL_DOWNLOAD",
        "USES_TERMINAL_UPDATE",
        "USES_TERMINAL_CONFIGURE",
        "USES_TERMINAL_BUILD",
        "USES_TERMINAL_INSTALL",
        "USES_TERMINAL_TEST",
        # Target
        "EXCLUDE_FROM_ALL",
        # Miscellaneous
        "LIST_SEPARATOR",
    ]
    multi_value_keywords = [
        # Download Step
        "DOWNLOAD_COMMAND",
        "URL",
        "HTTP_HEADER",
        "GIT_SUBMODULES",
        "GIT_CONFIG",
        # Update/Patch Step
        "UPDATE_COMMAND",
        "PATCH_COMMAND",
        # Configure Step
        "CONFIGURE_COMMAND",
        "CMAKE_ARGS",
        "CMAKE_CACHE_ARGS",
        "CMAKE_CACHE_DEFAULT_ARGS",
        # Build Step
        "BUILD_COMMAND",
        "BUILD_BYPRODUCTS",
        # Install Step
        "INSTALL_COMMAND",
        # Test Step
        "TEST_COMMAND",
        # Output Logging
        # Terminal Access
        # Target
        "DEPENDS",
        "STEP_TARGETS",
        "INDEPENDENT_STEP_TARGETS",
        # Miscellaneous
        "COMMAND",
    ]
    keyword_formatters = {
        key: "_format_command_line"
        for key in [
            "DOWNLOAD_COMMAND",
            "GIT_CONFIG",
            "UPDATE_COMMAND",
            "PATCH_COMMAND",
            "CONFIGURE_COMMAND",
            "CMAKE_ARGS",
            "CMAKE_CACHE_ARGS",
            "CMAKE_CACHE_DEFAULT_ARGS",
            "BUILD_COMMAND",
            "INSTALL_COMMAND",
            "TEST_COMMAND",
            "COMMAND",
        ]
    }


class ExternalProjectAddStep(
    CommandLineFormatter, ArgumentAwareCommandInvocationDumper
):
    one_value_keywords = [
        "COMMENT",
        "ALWAYS",
        "EXCLUDE_FROM_MAIN",
        "WORKING_DIRECTORY",
        "LOG",
        "USES_TERMINAL",
    ]
    multi_value_keywords = [
        "COMMAND",
        "DEPENDEES",
        "DEPENDERS",
        "DEPENDS",
        "BYPRODUCTS",
    ]
    keyword_formatters = {"COMMAND": "_format_command_line"}


class FeatureSummary(ArgumentAwareCommandInvocationDumper):
    options = [
        "APPEND",
        "INCLUDE_QUIET_PACKAGES",
        "FATAL_ON_MISSING_REQUIRED_PACKAGES",
        "QUIET_ON_EMPTY",
        "DEFAULT_DESCRIPTION",
    ]
    one_value_keywords = ["FILENAME", "VAR", "DESCRIPTION", "WHAT"]


class SetPackageProperties(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["URL", "DESCRIPTION", "TYPE", "PURPOSE"]


class FetchContentDeclare(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    one_value_keywords = [
        # Download Step
        "URL_HASH",
        "URL_MD5",
        "DOWNLOAD_NAME",
        "DOWNLOAD_NO_EXTRACT",
        "TIMEOUT",
        "HTTP_USERNAME",
        "HTTP_PASSWORD",
        "TLS_VERIFY",
        "TLS_CAINFO",
        "NETRC",
        "NETRC_FILE",
        "GIT_REPOSITORY",
        "GIT_TAG",
        "GIT_REMOTE_NAME",
        "GIT_SUBMODULES_RECURSE",
        "GIT_SHALLOW",
        "GIT_PROGRESS",
        "GIT_REMOTE_UPDATE_STRATEGY",
        "SVN_REPOSITORY",
        "SVN_REVISION",
        "SVN_USERNAME",
        "SVN_PASSWORD",
        "SVN_TRUST_CERT",
        "HG_REPOSITORY",
        "HG_TAG",
        "CVS_REPOSITORY",
        "CVS_MODULE",
        "CVS_TAG",
        "SOURCE_SUBDIR",
        # Update/Patch Step
        "UPDATE_DISCONNECTED",
    ]
    multi_value_keywords = [
        # Download Step
        "DOWNLOAD_COMMAND",
        "URL",
        "HTTP_HEADER",
        "GIT_SUBMODULES",
        "GIT_CONFIG",
        # Update/Patch Step
        "UPDATE_COMMAND",
        "PATCH_COMMAND",
    ]
    keyword_formatters = {
        key: "_format_command_line"
        for key in ["DOWNLOAD_COMMAND", "UPDATE_COMMAND", "PATCH_COMMAND"]
    }


class FetchContentPopulate(ArgumentAwareCommandInvocationDumper):
    options = ["QUIET"]
    one_value_keywords = [
        "SUBBUILD_DIR",
        "SOURCE_DIR",
        "BINARY_DIR",
        # Same as externalproject_add
        # Download Step
        "URL_HASH",
        "URL_MD5",
        "DOWNLOAD_NAME",
        "DOWNLOAD_NO_EXTRACT",
        "TIMEOUT",
        "HTTP_USERNAME",
        "HTTP_PASSWORD",
        "TLS_VERIFY",
        "TLS_CAINFO",
        "NETRC",
        "NETRC_FILE",
        "GIT_REPOSITORY",
        "GIT_TAG",
        "GIT_REMOTE_NAME",
        "GIT_SUBMODULES_RECURSE",
        "GIT_SHALLOW",
        "GIT_PROGRESS",
        "GIT_REMOTE_UPDATE_STRATEGY",
        "SVN_REPOSITORY",
        "SVN_REVISION",
        "SVN_USERNAME",
        "SVN_PASSWORD",
        "SVN_TRUST_CERT",
        "HG_REPOSITORY",
        "HG_TAG",
        "CVS_REPOSITORY",
        "CVS_MODULE",
        "CVS_TAG",
        # Update/Patch Step
        "UPDATE_DISCONNECTED",
    ]
    multi_value_keywords = [
        # Same as externalproject_add
        # Download Step
        "DOWNLOAD_COMMAND",
        "URL",
        "HTTP_HEADER",
        "GIT_SUBMODULES",
        "GIT_CONFIG",
        # Update/Patch Step
        "UPDATE_COMMAND",
        "PATCH_COMMAND",
    ]


class FetchContentGetProperties(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["SOURCE_DIR", "BINARY_DIR", "POPULATED"]


class FindPackageHandleStandardArgs(ArgumentAwareCommandInvocationDumper):
    options = ["HANDLE_COMPONENTS", "CONFIG_MODE", "NAME_MISMATCHED"]
    one_value_keywords = [
        "FOUND_VAR",
        "VERSION_VAR",
        "REASON_FAILURE_MESSAGE",
        "FAIL_MESSAGE",
    ]
    multi_value_keywords = ["REQUIRED_VARS"]


class FindPackageCheckVersion(ArgumentAwareCommandInvocationDumper):
    options = ["HANDLE_VERSION_RANGE"]
    one_value_keywords = ["RESULT_MESSAGE_VARIABLE"]


class FortranCInterfaceHeader(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["MACRO_NAMESPACE", "SYMBOL_NAMESPACE"]
    multi_value_keywords = ["SYMBOLS"]


class GenerateExportHeader(ArgumentAwareCommandInvocationDumper):
    options = ["DEFINE_NO_DEPRECATED"]
    one_value_keywords = [
        "BASE_NAME",
        "EXPORT_MACRO_NAME",
        "EXPORT_FILE_NAME",
        "DEPRECATED_MACRO_NAME",
        "NO_EXPORT_MACRO_NAME",
        "INCLUDE_GUARD_NAME",
        "STATIC_DEFINE",
        "NO_DEPRECATED_MACRO_NAME",
        "PREFIX_NAME",
        "CUSTOM_CONTENT_FROM_VARIABLE",
    ]


class GTestAddTests(ArgumentAwareCommandInvocationDumper):
    options = ["SKIP_DEPENDENCY"]
    one_value_keywords = [
        "TARGET",
        "WORKING_DIRECTORY",
        "TEST_PREFIX",
        "TEST_SUFFIX",
        "TEST_LIST",
    ]
    multi_value_keywords = ["SOURCES", "EXTRA_ARGS"]


class GTestDiscoverTests(
    CommandLineFormatter,
    KeywordWithPairsFormatter,
    ArgumentAwareCommandInvocationDumper,
):
    options = ["NO_PRETTY_TYPES", "NO_PRETTY_VALUES"]
    one_value_keywords = [
        "WORKING_DIRECTORY",
        "TEST_PREFIX",
        "TEST_SUFFIX",
        "TEST_LIST",
        "DISCOVERY_TIMEOUT",
        "XML_OUTPUT_DIR",
        "DISCOVERY_MODE",
        "TEST_FILTER",
    ]
    multi_value_keywords = ["EXTRA_ARGS", "PROPERTIES"]
    keyword_formatters = {
        "EXTRA_ARGS": "_format_command_line",
        "PROPERTIES": "_format_keyword_with_pairs",
    }


class AddJar(ArgumentAwareCommandInvocationDumper):
    options = ["RESOURCES"]
    one_value_keywords = [
        "ENTRY_POINT",
        "VERSION",
        "OUTPUT_NAME",
        "OUTPUT_DIR",
        "GENERATE_NATIVE_HEADERS",
        "DESTINATION",
        "BUILD",
    ]
    multi_value_keywords = ["SOURCES", "INCLUDE_JARS", "NAMESPACE"]


class InstallJar(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DESTINATION", "COMPONENT"]


class InstallJniSymlink(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DESTINATION", "COMPONENT"]


class InstallJarExports(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["NAMESPACE", "FILE", "DESTINATION", "COMPONENT"]
    multi_value_keywords = ["TARGETS"]


class ExportJars(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["NAMESPACE", "FILE"]
    multi_value_keywords = ["TARGETS"]


class FindJar(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DOC", "ENV"]
    multi_value_keywords = ["NAMES", "PATHS", "VERSIONS"]


class CreateJavaDoc(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = [
        "SOURCEPATH",
        "CLASSPATH",
        "INSTALLPATH",
        "DOCTITLE",
        "WINDOWTITLE",
        "AUTHOR",
        "USE",
        "VERSION",
    ]
    multi_value_keywords = ["PACKAGES", "FILES"]


class CreateJavah(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["TARGET", "GENERATED_FILES", "OUTPUT_NAME", "OUTPUT_DIR"]
    multi_value_keywords = ["CLASSES", "CLASSPATH", "DEPENDS"]


class SwigAddLibrary(ArgumentAwareCommandInvocationDumper):
    options = ["NO_PROXY"]
    one_value_keywords = ["TYPE", "LANGUAGE", "OUTPUT_DIR", "OUTFILE_DIR"]
    multi_value_keywords = ["SOURCES"]


class WriteCompilerDetectionHeader(ArgumentAwareCommandInvocationDumper):
    options = ["ALLOW_UNKNOWN_COMPILERS", "ALLOW_UNKNOWN_COMPILER_VERSIONS"]
    one_value_keywords = [
        "FILE",
        "PREFIX",
        "OUTPUT_FILES_VAR",
        "OUTPUT_DIR",
        "VERSION",
        "PROLOG",
        "EPILOG",
    ]
    multi_value_keywords = ["COMPILERS", "FEATURES", "BARE_FEATURES"]


class WriteBasicPackageVersionFile(ArgumentAwareCommandInvocationDumper):
    options = ["ARCH_INDEPENDENT"]
    one_value_keywords = ["VERSION", "COMPATIBILITY"]


class BisonTarget(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["COMPILE_FLAGS", "DEFINES_FILES", "VERBOSE", "REPORT_FILE"]


class DoxygenAddDocs(ArgumentAwareCommandInvocationDumper):
    options = ["ALL", "USE_STAMP_FILE"]
    one_value_keywords = ["WORKING_DIRECTORY", "COMMENT"]


class EnvModule(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["OUTPUT_VARIABLE", "RESULT_VARIABLE"]
    multi_value_keywords = ["COMMAND"]
    keyword_formatters = {"COMMAND": "_format_command_line"}


class EnvModuleSwap(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["OUTPUT_VARIABLE", "RESULT_VARIABLE"]


class FlexTarget(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["COMPILE_FLAGS", "DEFINES_FILES"]


class GettextCreateTranslations(ArgumentAwareCommandInvocationDumper):
    options = ["ALL"]


class GettextProcessPotFile(ArgumentAwareCommandInvocationDumper):
    options = ["ALL"]
    one_value_keywords = ["INSTALL_DESTINATION"]
    multi_value_keywords = ["LANGUAGES"]


class GettextProcessPoFiles(ArgumentAwareCommandInvocationDumper):
    options = ["ALL"]
    one_value_keywords = ["INSTALL_DESTINATION"]
    multi_value_keywords = ["PO_FILES"]


class MatlabAddUnitTest(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    options = ["NO_UNITTEST_FRAMEWORK"]
    one_value_keywords = [
        "NAME",
        "UNITTEST_FILE",
        "CUSTOM_TEST_COMMAND",
        "UNITTEST_PRECOMMAND",
        "TIMEOUT",
    ]
    multi_value_keywords = [
        "ADDITIONAL_PATH",
        "MATLAB_ADDITIONAL_STARTUP_OPTIONS",
        "TEST_ARGS",
    ]
    keyword_formatters = {"MATLAB_ADDITIONAL_STARTUP_OPTIONS": "_format_command_line"}


class MatlabAddMex(ArgumentAwareCommandInvocationDumper):
    options = ["EXECUTABLE", "MODULE", "SHARED", "R2017b", "R2018a", "EXCLUDE_FROM_ALL"]
    one_value_keywords = ["NAME", "OUTPUT_NAME", "DOCUMENTATION"]
    multi_value_keywords = ["SRC", "LINK_TO"]


class PkgCheckModules(ArgumentAwareCommandInvocationDumper):
    options = [
        "REQUIRED",
        "QUIET",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "IMPORTED_TARGET",
        "GLOBAL",
    ]


class PkgSearchModule(ArgumentAwareCommandInvocationDumper):
    options = [
        "REQUIRED",
        "QUIET",
        "NO_CMAKE_PATH",
        "NO_CMAKE_ENVIRONMENT_PATH",
        "IMPORTED_TARGET",
        "GLOBAL",
    ]


class ProtobufGenerateCpp(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["DESCRIPTORS", "EXPORT_MACRO"]


class Qt4WrapCpp(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["TARGET"]
    multi_value_keywords = ["OPTIONS"]


class Qt4WrapUi(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["OPTIONS"]


class Qt4AddResources(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["OPTIONS"]


class Qt4GenerateMoc(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["TARGET"]


class Qt4GenerateDbusInterface(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["OPTIONS"]


class Qt4CreateTranslation(ArgumentAwareCommandInvocationDumper):
    multi_value_keywords = ["OPTIONS"]


class Qt4Automoc(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["TARGET"]


class SquishAddTest(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = [
        "AUT",
        "SUITE",
        "TEST",
        "SETTINGSGROUP",
        "PRE_COMMAND",
        "POST_COMMAND",
    ]


class ExternalDataAddTarget(ArgumentAwareCommandInvocationDumper):
    one_value_keywords = ["SHOW_PROGRESS"]


module_command_mapping = {
    # Utility Modules
    "android_add_test_data": AndroidAddTestData,
    "check_c_source_compiles": CheckCSourceCompiles,
    "check_cxx_source_compiles": CheckCXXSourceCompiles,
    "check_fortran_source_compiles": CheckFortranSourceCompiles,
    "check_fortran_source_runs": CheckFortranSourceRuns,
    "check_include_files": CheckIncludeFiles,
    "check_ipo_supported": CheckIPOSupported,
    "check_objc_source_compiles": CheckOBJCSourceCompiles,
    "check_objcxx_source_compiles": CheckOBJCXXSourceCompiles,
    "check_pie_supported": CheckPIESupported,
    "check_struct_has_member": CheckStructHasMember,
    "check_type_size": CheckTypeSize,
    "cmake_add_fortran_subdirectory": CMakeAddFortranSubdirectory,
    "cmake_print_properties": CMakePrintProprties,
    "configure_package_config_file": ConfigurePackageConfigFile,
    "cpack_add_component": CPackAddComponent,
    "cpack_add_component_group": CPackAddComponentGroup,
    "cpack_add_install_type": CPackAddInstallType,
    "cpack_configure_downloads": CPackConfigureDownloads,
    "cpack_ifw_add_repository": CPackIFWAddRepository,
    "cpack_ifw_configure_component": CPackIFWConfigureComponent,
    "cpack_ifw_configure_component_group": CPackIFWConfigureComponentGroup,
    "cpack_ifw_update_repository": CPackIFWUpdateRepository,
    "ctest_coverage_collect_gcov": CTestCoverageCollectGCOV,
    "externalproject_add": ExternalProjectAdd,
    "externalproject_add_step": ExternalProjectAddStep,
    "externaldata_add_target": ExternalDataAddTarget,
    "feature_summary": FeatureSummary,
    "set_package_properties": SetPackageProperties,
    "fetchcontent_declare": FetchContentDeclare,
    "fetchcontent_populate": FetchContentPopulate,
    "fetchcontent_getproperties": FetchContentGetProperties,
    "find_package_handle_standard_args": FindPackageHandleStandardArgs,
    "find_package_check_version": FindPackageCheckVersion,
    "fortrancinterface_header": FortranCInterfaceHeader,
    "generate_export_header": GenerateExportHeader,
    "gtest_add_tests": GTestAddTests,
    "gtest_discover_tests": GTestDiscoverTests,
    "add_jar": AddJar,
    "install_jar": InstallJar,
    "install_jni_symlink": InstallJniSymlink,
    "install_jar_exports": InstallJarExports,
    "export_jars": ExportJars,
    "find_jar": FindJar,
    "create_javadoc": CreateJavaDoc,
    "create_javah": CreateJavah,
    "swig_add_library": SwigAddLibrary,
    "write_compiler_detection_header": WriteCompilerDetectionHeader,
    "write_basic_package_version_file": WriteBasicPackageVersionFile,
    # Find Modules
    "bison_target": BisonTarget,
    "doxygen_add_docs": DoxygenAddDocs,
    "env_module": EnvModule,
    "env_module_swap": EnvModuleSwap,
    "flex_target": FlexTarget,
    "gettext_create_translations": GettextCreateTranslations,
    "gettext_process_pot_file": GettextProcessPotFile,
    "gettext_process_po_files": GettextProcessPoFiles,
    "matlab_add_unit_test": MatlabAddUnitTest,
    "matlab_add_mex": MatlabAddMex,
    "pkg_check_modules": PkgCheckModules,
    "pkg_search_module": PkgSearchModule,
    "protobuf_generate_cpp": ProtobufGenerateCpp,
    "qt4_wrap_cpp": Qt4WrapCpp,
    "qt4_wrap_ui": Qt4WrapUi,
    "qt4_add_resources": Qt4AddResources,
    "qt4_generate_moc": Qt4GenerateMoc,
    "qt4_generate_dbus_interface": Qt4GenerateDbusInterface,
    "qt4_create_translation": Qt4CreateTranslation,
    "qt4_automoc": Qt4Automoc,
    "squish_add_test": SquishAddTest,
}
